from enum import Enum
from logging import getLogger

from NEBULA.core.injectionImpl import InjectionImpl


class ErrorTypes(Enum):
    # TODO test all of this
    """All Available Errortypes"""
    NORMAL = "_injectNormalError"
    BURST = "_injectBurstError"
    STUCKAT = "_injectStuckAtError"

    def __new__(cls, func_name):
        # Assign function to each enum instance
        obj = object.__new__(cls)
        obj._value_ = func_name
        obj.func = getattr(cls, func_name)
        obj._logger = getLogger(__name__)
        return obj

    def __call__(self, layername: str, layerMem: dict, probability: float):
        self._logger.debug(f"choosing strategy {self} for errorinjection")
        return self.func(layername, layerMem, probability)

    @staticmethod
    def _injectNormalError(layername: str, layerMem: dict, probability: float):
        return InjectionImpl._concurrentErrorInjection(layername, layerMem, probability)

    @staticmethod
    def _injectBurstError(layername: str, layerMem: dict, probability: float):
        return InjectionImpl._concurrentBurstInjection(layername, layerMem, probability)

    @staticmethod
    def _injectStuckAtError(layername: str, layerMem: dict, probability: float):
        raise NotImplementedError("Stuck at Error is not implemented yet")

    @staticmethod
    def _injectCustom(layername: str, layerMem: dict, probability: float):
        """
        used to enforce signature to ensure it works in multiprocessing env
        when user subclasses enum class to add own errortypes
        """
        pass



#!/usr/bin/env python3

"""
history.py
    provides utility to undo changes made by the errorinjector
"""

__author__      = "Alexander Tepe"
__email__       = "alexander.tepe@hotmail.de"
__copyright__   = "Copyright 2024, Planet Earth"

import copy
from collections import deque

from keras import Layer


class History(deque):
    """Subclass of the deque class
    acts like a stack for Model instances.
    Interface is as you would expect:
        push - add element to top of stack
        pop - get and remove from top of stack
        peek - get element from top without removing it
    """

    def __init__(self, layers: list[Layer] = []) -> None:
        super().__init__()
        if layers is not []:
            layerCopy = copy.deepcopy(layers)
            self.push(layerCopy)

    def push(self, entry: list[Layer]) -> None:
        layerCopy = copy.deepcopy(entry)
        self.append(layerCopy)

    def revert(self) -> None:
        """ Revert last change made to history
        Removes the element from top of the stack
        """
        try:
            super().pop()
        except IndexError:
            raise IndexError("pop from an empty history")

    def pop(self) -> list[Layer]:
        try:
            return super().pop()
        except IndexError:
            raise IndexError("pop from an empty history")

    def peek(self) -> list[Layer]:
        try:
            elem = super().pop()
            super().append(elem)
            return elem
        except IndexError:
            raise IndexError("peek from an empty history")

    def size(self):
        return len(self)



from NEBULA.core.legacy import flip_random_bits_in_model_weights, flip_single_number_float
from NEBULA.core.injector import Injector
from NEBULA.core.legacyInjector import LegacyInjector
from NEBULA.core.trainingInjector import TrainingInjector
from NEBULA.core.noiseLayer import NoiseLayer
from NEBULA.core.errorTypes import ErrorTypes

__all__ = [
    "flip_single_number_float",
    "flip_random_bits_in_model_weights",
    "Injector",
    "LegacyInjector",
    "TrainingInjector",
    "NoiseLayer",
    "ErrorTypes",
]



#!/usr/bin/env python3

"""
TrainingInjector.py
    injector to attach to untrained model to simulate errors during training
"""

__author__      = "Alexander Tepe"
__email__       = "alexander.tepe@hotmail.de"
__copyright__   = "Copyright 2024, Planet Earth"

from logging import Logger

from keras import Model, Layer
from keras.src.models import Functional, Sequential

from NEBULA.core.noiseLayer import NoiseLayer
from NEBULA.utils.logging import getLogger


def _buildSequentialModel(model: Model, noiseLayer: Layer, index: int) -> Model:
    layers = model.layers
    pre_layers = layers[:index]  # layers before the insertion point
    post_layers = layers[index:]  # layers after the insertion point

    newModel = Sequential()

    for layer in pre_layers:
        newModel.add(layer)
    newModel.add(noiseLayer)
    for layer in post_layers:
        newModel.add(layer)

    return newModel


def _buildFunctionalModel(model: Model, noiseLayer: Layer, index: int) -> Model:
    inputs = model.input
    x = inputs

    for i, layer in enumerate(model.layers):
        if i == index:
            x = noiseLayer(x)
        if i != 0:
            x = layer(x)

    return Functional(inputs=inputs, outputs=x)


class TrainingInjector():
    """Use to attach to untrained models
    This injector will simulate biterrors during the training phase
    by adding a noiselayer to the given model.
    This layer will produce noise during the learning phase of the network.
    During inference the layer will pass all values from the n-1th layer through
    """

    _logger: Logger
    _probability: float
    _clipping: tuple | None

    def __init__(self, probability: float = 0.01, clipping: tuple | None = None) -> None:
        if probability < .0:
            raise ValueError("BER cannot be negative")
        self._logger = getLogger(__name__)
        self._probability = probability
        self._clipping = clipping

    def attach(self, model: Model, index: int = 1) -> Model:
        """Attach error injecting layer to existing untrained model
        This method will insert a layer into the given model to
        cause noise during backpropagation in the training of the model.
        index parameter specifies where in the model to attach the pertubating layer.
        The placement has severe impact on the trained model
        """
        if index < 0:
            raise ValueError("Index cannot be negative")
        nl = NoiseLayer(probability=self._probability, clipping=self._clipping)

        if isinstance(model, Sequential):
            return _buildSequentialModel(model, nl, index)

        elif isinstance(model, Functional):
            return _buildFunctionalModel(model, nl, index)

        else:
            raise ValueError("Type of model is not supported")



import unittest

import numpy as np
import tensorflow as tf
from tensorflow._api.v2.errors import InvalidArgumentError

from NEBULA.core.noiseLayer import NoiseLayer


class NoiseLayerTest(unittest.TestCase):

    def test_callShouldInjectErrors(self):
        with self.assertRaises(InvalidArgumentError):
            inputs = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0], dtype=np.float32)
            inputs = tf.constant(inputs)
            nl = NoiseLayer(probability=1.0)
            corruptedInputs = nl.call(inputs, training=True)
            tf.debugging.assert_equal(inputs, corruptedInputs)

    def test_callShouldNotInjectDuringInference(self):
        try:
            inputs = np.array([1.0, 1.0, 1.0, 1.0, 1.0, 1.0], dtype=np.float32)
            inputs = tf.constant(inputs)
            nl = NoiseLayer(probability=1.0)
            corruptedInputs = nl.call(inputs, training=False)
            tf.debugging.assert_equal(inputs, corruptedInputs)
        except InvalidArgumentError:
            self.fail()

    def test_invalidClippingParseDiesOnInvalidTuple(self):
        with self.assertRaises(ValueError):
            _ = NoiseLayer(probability=0.0, clipping=(100, 0))

    def test_wrongTypeInClippingRaisesError(self):
        with self.assertRaises(ValueError):
            _ = NoiseLayer(clipping=("string", "NotANumber"))

    def test_setClippingTupleWorks(self):
        nl = NoiseLayer(probability=0.5, clipping=(-1, 1))
        self.assertEqual(nl.probability, 0.5)
        self.assertEqual(nl.clipping, (-1, 1))

    def test_clippingIsUsed(self):
        inputs = np.array([10.0, 10.0, -10.0, 10.0, -10.0, 10.0], dtype=np.float32)
        inputs = tf.constant(inputs)

        nl = NoiseLayer(probability=0.0, clipping=(-1.0, 1.0))
        res = nl.call(inputs, training=True)
        for item in res.numpy():
            self.assertGreaterEqual(item, -1.0)
            self.assertLessEqual(item, 1.0)



import unittest
from multiprocessing import shared_memory
from unittest.mock import Mock

import numpy as np

from NEBULA.core.injectionImpl import InjectionImpl
from NEBULA.utils.logging import getLogger


class TestInjectorImpl(unittest.TestCase):

    _model = None
    _layerMem = {
        "membuf": [None],
        "shapes": [(2,)]
    }
    _logger = getLogger(__name__)

    def setUp(self):
        self._model = Mock()
        self._model.get_weights.return_value = [1, 2]
        data = np.array([1, 2])
        self.shm = shared_memory.SharedMemory(create=True, size=data.nbytes)
        sharedData = np.ndarray(data.shape, dtype=data.dtype, buffer=self.shm.buf)
        np.copyto(sharedData, data)
        self._layerMem["membuf"][0] = self.shm

    def tearDown(self):
        try:
            self._layerMem["membuf"][0].close()
            self._layerMem["membuf"][0].unlink()
        except KeyError:
            self._logger.warn("Could not clean up shared memory from unittest")

    def test_ConcurrentRoutine(self):
        origWeights = self._model.get_weights()

        layerName, newWeights = InjectionImpl._concurrentErrorInjection("Test", self._layerMem, probability=1.0)

        self.assertEqual("Test", layerName)
        self.assertNotEqual(origWeights, newWeights)

    def test_concurrentBurstRoutine(self):
        origWeights = self._model.get_weights()

        layerName, newWeights = InjectionImpl._concurrentBurstInjection("Test", self._layerMem, probability=1.0)

        self.assertEqual("Test", layerName)
        self.assertNotEqual(origWeights, newWeights)



#!/usr/bin/env python3

"""
logging.py:
    All logging related methods for API internal logging
"""

__author__      = "Alexander Tepe"
__email__       = "alexander.tepe@hotmail.de"
__copyright__   = "Copyright 2024, Planet Earth"

import logging
import os

from dotenv import load_dotenv

# init logger once when lib is imported first
# logging not exposed to the user on purpose

_loggers = {}

load_dotenv()
_LOG_LEVEL = logging.INFO

try:
    level = os.environ["NEBULA_LOG_LEVEL"].upper()
    if level == "DEBUG":
        _LOG_LEVEL = logging.DEBUG
    elif level == "INFO":
        _LOG_LEVEL = logging.INFO
    elif level == "WARNING":
        _LOG_LEVEL = logging.WARNING
    elif level == "ERROR":
        _LOG_LEVEL = logging.ERROR
    elif level == "CRITICAL":
        _LOG_LEVEL = logging.CRITICAL
except KeyError as e:
    print(f"Could not map log level from environment: {e.args[0]}")

logging.basicConfig(
    level=_LOG_LEVEL,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)


def getLogger(name: str) -> logging.Logger:
    """Get a logger instance with the given modulename
    Only one logger per modulename will be assigned a handler
    this way, the amount of resources consumed for libray imports stays slim
    """

    if name not in _loggers:
        logger = logging.getLogger(name)
        logger.setLevel(_LOG_LEVEL)

        if not logger.hasHandlers():
            console_handler = logging.StreamHandler()
            # console_handler.setLevel(_LOG_LEVEL)

            logger.addHandler(console_handler)

        _loggers[name] = logger

    return _loggers[name]


def setLoggingLevel(logLevel: int | str, name: str | None = None) -> None:
    """change level of logger with given name
    The log level determines the minimum level of messages to be logged
    possible values are (increasing severity):
        DEBUG
        INFO
        ERROR
        CRITICAL
    if name param is omitted, all loggers with assigned handlers will be set to specified level
    """
    if name is None:
        for logger in _loggers.values():
            logger.setLevel(logLevel)
    else:
        if name in _loggers:
            _loggers[name].setLevel(logLevel)



import unittest

import keras
import numpy as np

from NEBULA.core.legacyInjector import LegacyInjector


class LegacyInjectorTest(unittest.TestCase):

    _model = None

    def setUp(self):
        if self._model is None:
            inputs = keras.Input(shape=(37,))
            x = keras.layers.Dense(32, activation="relu")(inputs)
            outputs = keras.layers.Dense(5, activation="softmax")(x)
            self._model = keras.Model(inputs=inputs, outputs=outputs)

    def test_injectErrorWith0ProbabilityDoesNotChangeModel(self):
        li = LegacyInjector(self._model.layers, probability=0.0)
        weightsOld = self._model.get_weights()
        li.injectError(self._model)
        weightsNew = self._model.get_weights()
        for orig, new in zip(weightsNew, weightsOld):
            self.assertTrue(np.allclose(orig, new))

    def test_injectErrorDoesChangeModel(self):
        li = LegacyInjector(self._model.layers, probability=1.0)
        weightsOld = self._model.get_weights()
        li.injectError(self._model)
        weightsNew = self._model.get_weights()
        allSame = True
        for orig, new in zip(weightsOld, weightsNew):
            allSame = np.allclose(orig, new)
            if not allSame:
                break

        self.assertFalse(allSame)


import unittest

import keras
import numpy as np
from tensorflow import float32, Variable

from NEBULA.core.legacy import flip_random_bits_in_model_weights, flip_single_number_float


class LegacyTest(unittest.TestCase):

    def test_flipSingleNumberFloat(self):
        number = Variable(initial_value=123.0, dtype=float32)
        numberAltered = flip_single_number_float(number, probability=1.0)
        self.assertNotEqual(number, numberAltered)


    def test_flipRandomBitsInModelWeights(self):
        inputs = keras.Input(shape=(37,))
        x = keras.layers.Dense(32, activation="relu")(inputs)
        outputs = keras.layers.Dense(5, activation="softmax")(x)
        model = keras.Model(inputs=inputs, outputs=outputs)

        modelAltered = flip_random_bits_in_model_weights(model, probability=1.0)
        changed = False
        for new, orig in zip(modelAltered.get_weights(), model.get_weights()):
            changed = np.allclose(orig, new)
            if changed:
                break

        self.assertTrue(changed)



from NEBULA.utils.logging import setLoggingLevel
from NEBULA.core.legacyInjector import LegacyInjector
from NEBULA.core.injector import Injector
from NEBULA.core.trainingInjector import TrainingInjector
from NEBULA.utils.helper import loadFatModel, binary
from NEBULA.core.errorTypes import ErrorTypes

__all__ = [
    "setLoggingLevel",
    "LegacyInjector",
    "Injector",
    "TrainingInjector",
    "loadFatModel",
    "ErrorTypes",
    "binary",
]



import unittest
from unittest.mock import Mock

from NEBULA.core.history import History


class HistoryTest(unittest.TestCase):

    def setUp(self):
        # Create a mock Keras model
        self.mockLayer1 = Mock()
        self.mockLayer2 = Mock()
        self.mockLayer3 = Mock()
        # Set up the mock model's behavior
        self.mockLayer1.get_weights.return_value = 1
        self.mockLayer2.get_weights.return_value = 2
        self.mockLayer3.get_weights.return_value = 3

    def test_pushWorks(self):
        history = History()
        history.push(self.mockLayer1)
        result = history.pop()
        self.assertEqual(1, result.get_weights())

    def test_popWorks(self):
        history = History()
        history.push(self.mockLayer1)
        size1 = history.size()
        _ = history.pop()
        size2 = history.size()
        self.assertNotEqual(size1, size2)

    def test_isFifo(self):
        history = History()
        history.push(self.mockLayer1)
        history.push(self.mockLayer2)
        history.push(self.mockLayer3)

        first = history.pop()
        second = history.pop()
        third = history.pop()

        self.assertEqual(first.get_weights(), 3)
        self.assertEqual(second.get_weights(), 2)
        self.assertEqual(third.get_weights(), 1)

    def test_peekDoesNotDelete(self):
        history = History()
        history.push(self.mockLayer1)
        size1 = history.size()
        _ = history.peek()
        size2 = history.size()
        self.assertEqual(size1, size2)

    def test_revertWorks(self):
        history = History()
        history.push(self.mockLayer1)
        history.push(self.mockLayer2)
        layersAltered = history.peek()
        self.assertEqual(layersAltered.get_weights(), 2)
        history.revert()
        layers = history.peek()
        self.assertEqual(layers.get_weights(), 1)



import unittest
import logging
from NEBULA.utils.logging import getLogger, setLoggingLevel


class LoggingTest(unittest.TestCase):

    logger = getLogger(__name__)

    def test_getLogger(self):
        # logger should exist
        self.assertIsNotNone(LoggingTest.logger)
        level = LoggingTest.logger.level

    def test_setLevel(self):
        setLoggingLevel(logging.ERROR)
        self.assertEqual(LoggingTest.logger.level, logging.ERROR)
        setLoggingLevel(logging.INFO)

    def test_setLevelSpecific(self):
        setLoggingLevel(logging.ERROR, __name__)
        self.assertEqual(LoggingTest.logger.level, logging.ERROR)

    def test_loggerInjectedOnlyOnce(self):
        from NEBULA.utils.logging import _loggers
        loggerCounter1 = len(_loggers)
        getLogger(__name__)
        loggerCounter2 = len(_loggers)
        self.assertEqual(loggerCounter1, loggerCounter2)



#!/usr/bin/env python3

"""
injector.py
    injector to use multiprocessing to inject errors
"""

__author__      = "Alexander Tepe"
__email__       = "alexander.tepe@hotmail.de"
__copyright__   = "Copyright 2024, Planet Earth"

import multiprocessing as mp
from logging import Logger
from multiprocessing import shared_memory

import numpy as np
from keras import Model, Layer

from NEBULA.core.baseInjector import BaseInjector
from NEBULA.core.errorTypes import ErrorTypes
from NEBULA.utils.logging import getLogger


def _initialize_shared_weights(layers: list[Layer]) -> dict:
    """Initialize shared memory for each layer's weights."""
    shared_weights = {}
    for layer in layers:
        layer_name = layer.name
        # TODO probably add dtype in case we're getting funky with quantization
        shared_weights[layer_name] = {"membuf": [], "shapes": []}

        for weight in layer.get_weights():
            shared_mem = shared_memory.SharedMemory(create=True, size=weight.nbytes)
            shared_weight = np.ndarray(weight.shape, dtype=weight.dtype, buffer=shared_mem.buf)
            np.copyto(shared_weight, weight)

            shared_weights[layer_name]["membuf"].append(shared_mem)
            shared_weights[layer_name]["shapes"].append(weight.shape)

    return shared_weights


def _create_process_pool(layers: list[Layer]) -> mp.Pool:
    """Create a process pool with one process per layer."""
    num_processes = len(layers)
    return mp.Pool(num_processes)


class Injector(BaseInjector):
    """Class Injector:
    encapsulates all injection and other comfort functions towards
    modifying a model
    The injector will create a processpool at instantiation with one process
    per layer of the given model. These are used to inject errors into the model.
    This class also yields access to the history of changes made to the model through
    error injection
    """

    _logger: Logger
    _process_pool: mp.Pool = None
    _sharedWeights: dict

    def __init__(self, layers: list[Layer], probability: float = 0.01) -> None:
        super().__init__(layers, probability)
        self._logger = getLogger(__name__)

        if mp.current_process().name == 'MainProcess':
            self._sharedWeights = _initialize_shared_weights(layers)
            self._process_pool = _create_process_pool(layers)

    def __del__(self):
        self._logger.debug("Closing Process Pool and deleting shared memory")
        if self._process_pool is not None:
            self._process_pool.close()
            self._process_pool.terminate()
        self._deleteShareMem()

    def injectError(self, model: Model, errorType: ErrorTypes = ErrorTypes.NORMAL) -> None:
        """ Method to inject errors into the model
        This method edits the model in place!
        Uses one process per layer of the given model and injects biterrors into the model
        with a Bit Error Rate of the given probability.
        This method applies the given errortype to all layers of the model.
        It is possible to override this method to use different errortypes per modellayer
        """
        # TODO test Errortypes
        self._logger.debug(f"Injecting error with probability of {self._probability}")

        # inject error
        results = self._injectToWeights(errorType)
        self._reconstructModel(model, results)
        self._history.push(model.layers)
        self._deleteShareMem()
        self._sharedWeights = _initialize_shared_weights(model.layers)

    def _injectToWeights(self, errorType: ErrorTypes = ErrorTypes.NORMAL) -> dict:
        """Modify weights of model using multiprocessing.
        Tensorflow locks GIL which blocks all threads which are not tensorflow
        control flow. Processes can still run.
        Since python parameters are passed as object references, the dictionary is
        modified in place.
        This also applies the special errortype strategy given by the callable enum value
        """
        # Apply the error injection function to each layer in parallel
        results = self._process_pool.starmap_async(
            errorType,  # this is basically a function
            [(layer, self._sharedWeights[layer], self._probability) for layer in self._sharedWeights.keys()]
        )
        return results.get()

    def undo(self, model: Model) -> None:
        """Undo change made by injecting error into weight"""
        try:
            super().undo(model)
        except ValueError:
            raise (ValueError("You probably meant to pass in a different model"))

    def _deleteShareMem(self) -> None:
        """Helper function to securely delete shared memory
        """
        try:
            for layer in self._sharedWeights:
                for i in range(len(self._sharedWeights[layer]["membuf"])):
                    self._sharedWeights[layer]["membuf"][i].close()
                    self._sharedWeights[layer]["membuf"][i].unlink()
        except IndexError:
            self._logger.warning("Mismatch in memory allocation during deletion")
            pass



#!/usr/bin/env python3

"""
legacy.py:
    Original bit flip operations from the WSA example
    These implementations were left in to allow for benchmarking
"""

__author__      = "Tim Düe"

import numpy as np
import tensorflow as tf

def flip_single_number_float(number_to_flip, data_type = 32, probability = 0.001, check = -1):
    random_numbers = np.random.rand(data_type + 1)
    flipped_bit_positions = np.where(random_numbers < probability)[0]
    if flipped_bit_positions.size == 0:
        return number_to_flip

    for pos in flipped_bit_positions:
        if data_type == 32:
            flip_mask = tf.bitwise.left_shift(tf.cast(1, tf.int32), pos)
            bitcast_to_int32 = tf.bitcast(number_to_flip, tf.int32)
            flipped_value = tf.bitwise.bitwise_xor(flip_mask, bitcast_to_int32)
            bitcast_to_float = tf.bitcast(flipped_value, tf.float32)
        elif data_type == 16:
            flip_mask = tf.bitwise.left_shift(tf.cast(1, tf.int16), pos)
            bitcast_to_int16 = tf.bitcast(number_to_flip, tf.int16)
            flipped_value = tf.bitwise.bitwise_xor(flip_mask, bitcast_to_int16)
            bitcast_to_float = tf.bitcast(flipped_value, tf.float16)
        else:
            print("data type ", data_type, " not valid")
        number_to_flip = bitcast_to_float

    if abs(bitcast_to_float) > check and check != -1:
        return 0
    else:
        return bitcast_to_float

def flip_random_bits_in_model_weights(model, probability = 0.001, check=-1):
    for layer in model.layers:
        layer_weights = layer.get_weights()
        new_weights = []
        layer_idx = 1
        for weight_tensor in layer_weights:
            # print("Working on layer ", layer_idx, " out of ", len(layer_weights))
            layer_idx = layer_idx + 1
            if weight_tensor.dtype == np.float32:
                shape = weight_tensor.shape
                flattened_weights = weight_tensor.flatten()
                for i in range(len(flattened_weights)):
                    flattened_weights[i] = flip_single_number_float(flattened_weights[i], probability=probability, check=check)
                new_weight_tensor = flattened_weights.reshape(shape)
                new_weights.append(new_weight_tensor)
            else:
                new_weights.append(weight_tensor)
        layer.set_weights(new_weights)
    return model



import struct

from NEBULA.core.noiseLayer import NoiseLayer

from tensorflow.keras.models import load_model


def binary(num):
    """
    Helper Function to display the binary representation of floating point numbers
    Source: https://stackoverflow.com/questions/16444726/binary-representation-of-float-in-python-bits-not-hex
    Validated using https://www.h-schmidt.net/FloatConverter/IEEE754.html
    """
    return ''.join('{:0>8b}'.format(c) for c in struct.pack('!f', num))


def loadFatModel(path: str, *args, **kwargs):
    """Load Fault-Aware-Trained model using the NEBULA.traininginjector from .h5 file
    This passes all arguments through to the load_model function from keras,
    while supplying the underlying NoiseLayer NEBULA uses for fault aware training.
    This layer must be known for Keras in order to load the model correctly.
    """
    if 'custom_objects' not in kwargs:
        kwargs['custom_objects'] = {'NoiseLayer': NoiseLayer}
    return load_model(path, *args, **kwargs)



import tensorflow as tf

import numpy as np
import struct
import random


def flipFloat(number_to_flip, data_type=32, probability=0.001, check=-1):
    """Helper function which flips bits in a given memory range with a given probability
    returns the modified float number as a tf.tensor
    """
    random_numbers = np.random.rand(data_type + 1)
    flipped_bit_positions = np.where(random_numbers < probability)[0]
    if flipped_bit_positions.size == 0:
        return number_to_flip

    for pos in flipped_bit_positions:
        if data_type == 32:
            flip_mask = tf.bitwise.left_shift(tf.cast(1, tf.int32), pos)
            bitcast_to_int32 = tf.bitcast(number_to_flip, tf.int32)
            flipped_value = tf.bitwise.bitwise_xor(flip_mask, bitcast_to_int32)
            bitcast_to_float = tf.bitcast(flipped_value, tf.float32)
        elif data_type == 16:
            flip_mask = tf.bitwise.left_shift(tf.cast(1, tf.int16), pos)
            bitcast_to_int16 = tf.bitcast(number_to_flip, tf.int16)
            flipped_value = tf.bitwise.bitwise_xor(flip_mask, bitcast_to_int16)
            bitcast_to_float = tf.bitcast(flipped_value, tf.float16)
        else:
            print("data type ", data_type, " not valid")
        number_to_flip = bitcast_to_float

    if abs(bitcast_to_float) > check and check != -1:
        return 0
    else:
        return bitcast_to_float


def flipAdjacentBits(value: float, burstLength: int, probability: float) -> float:
    """
    Flips `n_bits` adjacent bits in a `float32` value at a random starting position,
    based on a given probability.

    Parameters:
        value (float): The original float32 value.
        n_bits (int): The number of adjacent bits to flip.
        probability (float): The probability (0 to 1) that the group of bits is flipped.

    Returns:
        float: The modified float32 value after flipping bits.
    """

    # Convert float32 to 32-bit binary representation
    packed = struct.pack('>f', value)
    int_value = struct.unpack('>I', packed)[0]

    # Convert to binary list of bits
    bit_list = list(f'{int_value:032b}')

    # Choose a random starting index for adjacent bits
    start_index = random.randint(0, 32 - burstLength)

    # Flip adjacent bits based on the probability
    if random.random() < probability:
        for i in range(start_index, start_index + burstLength):
            bit_list[i] = '1' if bit_list[i] == '0' else '0'

    # Convert back to integer
    flipped_int_value = int("".join(bit_list), 2)

    # Convert back to float32
    flipped_packed = struct.pack('>I', flipped_int_value)
    flipped_value = struct.unpack('>f', flipped_packed)[0]

    return flipped_value


def flipTensorBits(input: tf.Tensor, probability: float, dtype: np.dtype) -> tf.Tensor:
    if dtype is np.float32:
        x_bits = tf.bitcast(input, tf.int32)
        randomValues = tf.random.uniform(shape=tf.shape(x_bits), minval=0.0, maxval=1.0)
        flipMask = randomValues < probability
        bitPositions = tf.random.uniform(shape=tf.shape(x_bits), minval=0, maxval=32, dtype=tf.int32)
        bitFlips = tf.bitwise.left_shift(tf.ones_like(x_bits, dtype=tf.int32), bitPositions)

        flippedBits = tf.bitwise.bitwise_xor(x_bits, tf.where(flipMask, bitFlips, 0))
        flippedFloat = tf.bitcast(flippedBits, tf.float32)

        return flippedFloat
    else:
        return input



from NEBULA.utils.logging import setLoggingLevel, getLogger
from NEBULA.utils.commons import flipFloat, flipTensorBits
from NEBULA.utils.helper import loadFatModel, binary

__all__ = [
    "setLoggingLevel",
    "getLogger",
    "flipFloat",
    "flipTensorBits",
    "loadFatModel",
    "binary"
]



#!/usr/bin/env python3

"""
InjectionImpl.py
    Actual modification of model weights is done here
"""

__author__      = "Alexander Tepe"
__email__       = "alexander.tepe@hotmail.de"
__copyright__   = "Copyright 2024, Planet Earth"

from threading import get_ident
from functools import wraps

import numpy as np

from NEBULA.utils.commons import flipAdjacentBits, flipFloat
from NEBULA.utils.logging import getLogger


def handleShmError(func):
    # TODO test this
    """
    Errorhandler that wraps annotated functions
    In case of errors when parsing the shared memory this wrapper will handle the error
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyError as e:
            layername = kwargs.get("layername", "Unknown Layer")
            InjectionImpl._logger.error(f"Cannot access argument {e.args[0]} of shared memory layer {layername}")
            return layername, []
    return wrapper


class InjectionImpl:
    """Implementation of bit error injection to weights
    Since Tensorflow sets the GIL Lock for threads that reference model memory,
    this implementation uses processes

    Idea: Create deep copy of model before calling injection functions.
    Since processes operate on their own memory, one process per layer can modify
    the model's weights. When all processes are done, the model is written back
    """

    _logger = getLogger(__name__)

    @handleShmError
    @staticmethod
    def _concurrentErrorInjection(layername: str, layerMem: dict, probability: float):
        """Routine which is executed by the subprocesses
        The weights from the model's layer are read from shared memory
        and modified with a given probability and
        returns the modified weights.
        """
        InjectionImpl._logger.debug(
            f"started worker process {get_ident()} on layer {layername} with BER of {probability}"
        )

        weights = InjectionImpl._shmHelper(layerMem)
        newWeights = []
        for weight in weights:
            shape = weight.shape
            if weight.dtype == np.float32:
                flattenedWeights = weight.flatten()
                for i in range(len(flattenedWeights)):
                    flattenedWeights[i] = flipFloat(flattenedWeights[i], probability=probability)
                newWeight = flattenedWeights.reshape(shape)
                newWeights.append(newWeight)
            else:
                newWeights.append(weight)
        return layername, newWeights

    @staticmethod
    def _concurrentStuckAtInjection(layername: str, layerMem: dict, probability: float):
        # TODO implement this? Or not?
        return {}

    @handleShmError
    @staticmethod
    def _concurrentBurstInjection(layername: str, layerMem: dict, probability: float):
        InjectionImpl._logger.debug(
            f"started worker process {get_ident()} on layer {layername} with BER of {probability}"
        )

        weights = InjectionImpl._shmHelper(layerMem)
        newWeights = []

        for weight in weights:
            shape = weight.shape
            if weight.dtype == np.float32:
                flattenedWeights = weight.flatten()
                for i in range(len(flattenedWeights)):
                    flattenedWeights[i] = flipAdjacentBits(flattenedWeights[i], 3, probability)
                newWeight = flattenedWeights.reshape(shape)
                newWeights.append(newWeight)
            else:
                newWeights.append(weight)

        return layername, newWeights

    @staticmethod
    def _shmHelper(layerMem: dict) -> list[np.ndarray]:
        weights = []
        for shm, shape in zip(layerMem["membuf"], layerMem["shapes"]):
            # TODO do not hardcode Float32!!!!
            weights.append(np.ndarray(shape, dtype=np.float32, buffer=shm.buf))
        return weights



#!/usr/bin/env python3

"""
BaseInjector.py:
    base class for all injectors
"""

__author__      = "Alexander Tepe"
__email__       = "alexander.tepe@hotmail.de"
__copyright__   = "Copyright 2024, Planet Earth"

from abc import ABC, abstractmethod

from keras import Model

from NEBULA.core.history import History


class BaseInjector(ABC):
    """Abstract base class for all injectors
    Injectors can be configured using the setter methods
    """

    _layers = None
    _probability = 0.01
    _history: History

    def __init__(
            self,
            layers,
            probability=_probability,
    ):
        self._layers = layers
        self._probability = probability
        self._history = History(self._layers)

    @abstractmethod
    def injectError(self, model) -> None:
        """Inject Errors into network
        Every subclass of the BaseInjector must implement this method
        """
        pass

    def undo(self, model: Model) -> None:
        """resets the last changes made to the model by the injector
        this will modify the model given by the param model and does not return anything.
        Will raise an ValueError if the history's layers cannot be written back into the specified model
        """
        self._history.revert()
        layers = self._history.peek()
        for layer in layers:
            model.get_layer(name=layer.name).set_weights(layer.get_weights())

    def _reconstructModel(self, model, result: dict) -> None:
        for item in result:
            if len(item[1]) > 0:
                model.get_layer(name=item[0]).set_weights(item[1])

    @property
    def layers(self):
        return self._layers

    @property
    def probability(self):
        return self._probability

    @layers.setter
    def layers(self, model):
        self._layers = model

    @probability.setter
    def probability(self, probability):
        self._probability = probability



import copy
import unittest

from keras import Model
import numpy as np

from NEBULA.core.injector import Injector
from NEBULA.utils.logging import getLogger
from NEBULA.core.errorTypes import ErrorTypes
from utils.ModelUtils import ModelUtils

import tensorflow.keras


class InjectorTest(unittest.TestCase):

    _model = None
    _logger = getLogger(__name__)

    def setUp(self):
        if self._model is None:
            self._model = ModelUtils.getBasicModel()

    def test_injectorWithHundredProbChangesModel(self):
        weightsOrig = self._model.get_weights()
        injector = Injector(self._model.layers, probability=1.0)
        injector.injectError(self._model)
        weightsNew = self._model.get_weights()

        allSame = True
        for orig, new in zip(weightsOrig, weightsNew):
            allSame = np.allclose(orig, new)
            if not allSame:
                break
        self.assertFalse(allSame)

    def test_reconstructModelWorks(self):
        injector = Injector(self._model.layers)
        modelCopy = ModelUtils.getBasicModel()

        # overwrite copy weights with zeros
        for idx, layer in enumerate(self._model.layers):
            original_weights = layer.get_weights()
            new_weights = [np.zeros(w.shape) for w in original_weights]
            copyLayer = modelCopy.get_layer(index=idx)
            copyLayer.set_weights(new_weights)

        # bring layer objects into dict form
        layers = list()
        for idx, layer in enumerate(modelCopy.layers):
            # Get the name of the layer
            layer_name = layer.name
            layer_weights = self._model.get_layer(index=idx).get_weights()
            # Store the weights in the dictionary
            layers.append((layer_name, layer_weights))

        # self._logger.debug(layers)

        injector._reconstructModel(modelCopy, layers)

        for orig, new in zip(self._model.get_weights(), modelCopy.get_weights()):
            self.assertTrue(np.allclose(orig, new))

    def test_processPoolCanExecuteMultipleTasks(self):
        injector = Injector(self._model.layers, probability=0.000005)
        try:
            injector.injectError(self._model)
            injector.injectError(self._model)
        except ValueError:
            # ValueError is thrown when task is queued before last injection is done
            self.fail()


    def test_changesAreReversible(self):
        weightsOrig = self._model.get_weights()

        #inject error
        injector = Injector(self._model.layers, probability=1.0)
        injector.injectError(self._model)

        weightsChanged = self._model.get_weights()

        allSame = True
        for orig, new in zip(weightsOrig, weightsChanged):
            allSame = np.allclose(orig, new)
            if not allSame:
                break
        self.assertFalse(allSame)

        injector.undo(self._model)

        weightsUndone = self._model.get_weights()

        for orig, new in zip(weightsOrig, weightsUndone):
            self.assertTrue(np.allclose(orig, new))

    def test_undoOnEmptyHistoryShouldRaiseIndexError(self):
        with self.assertRaises(IndexError):
            injector = Injector(self._model.layers)
            injector.undo(self._model)

    def test_undoOnWrongModelShouldRaiseAttributeError(self):
        with self.assertRaises(ValueError):
            injector = Injector(self._model.layers, probability=.0)
            otherModel = ModelUtils.getBasicModel()
            injector.injectError(self._model)
            injector.undo(otherModel)

    def test_updatingAfterInjectionWorks(self):
        injector = Injector(self._model.layers, probability=1.0)

        for layer in self._model.layers:
            data = injector._sharedWeights[layer.name]
            for idx, weight in enumerate(layer.get_weights()):
                weightFromBuffer = np.ndarray(
                    data['shapes'][idx],
                    dtype=np.float32,
                    buffer=data['membuf'][idx].buf
                )
                if np.any(np.isnan(weight)):
                    self.assertEqual(len(weight), len(weightFromBuffer))
                else:
                    self.assertTrue(np.allclose(weight, weightFromBuffer))

    def test_injectorWorksWithBurstErrorType(self):
        weightsOrig = self._model.get_weights()
        injector = Injector(self._model.layers, probability=1.0)
        injector.injectError(self._model, errorType=ErrorTypes.BURST)
        weightsNew = self._model.get_weights()

        allSame = True
        for orig, new in zip(weightsOrig, weightsNew):
            allSame = np.allclose(orig, new)
            if not allSame:
                break
        self.assertFalse(allSame)



#!/usr/bin/env python3

"""
noiseLayer.py
    Class to mimic noise during training
"""

__author__      = "Alexander Tepe"
__email__       = "alexander.tepe@hotmail.de"
__copyright__   = "Copyright 2024, Planet Earth"

from logging import Logger

import numpy as np
import tensorflow as tf
from keras.src import Layer

from NEBULA.utils.commons import flipTensorBits
from NEBULA.utils.logging import getLogger


class NoiseLayer(Layer):
    """subclass of keras layer simulating noise during training
    This class is a subclass of keras.layers.layer
    and can be used as such.
    It can be added to any model and will take the shape of the previous layer
    During training, the weights will add noise to the network during forwardpropagation
    """

    _logger: Logger
    _errorProbability: float
    _clipping: None | tuple = None

    def __init__(self, probability: float = 0.01, clipping: tuple | None = None, **kwargs):
        super().__init__(**kwargs)
        self.trainable = False
        self._logger = getLogger(__name__)
        self._errorProbability = probability
        self._parseClipping(clipping)

    def call(self, inputs, training=None):
        """Injects noise into model during training
        This method is called by keras during traning.
        While feeding the data through the model (before evaluating the loss function)
        this will take the values from the preceeding layers and modfiy them with a given probability.
        This will pertubate the results of the model during training.
        """
        if training is True:
            self._logger.debug(f"injecting errors during training with BER of {self._errorProbability}")
            results = tf.map_fn(self._outerHelper, inputs)
            if self._clipping is not None:
                self._logger.debug(f"clipping enabled with min: {self._clipping[0]}, min: {self._clipping[1]}")
                results = tf.clip_by_value(results, self._clipping[0], self._clipping[1])
            return results

        return inputs  # During inference, no noise is added

    def _outerHelper(self, x):
        return flipTensorBits(x, probability=self._errorProbability, dtype=np.float32)

    def _parseClipping(self, clipping) -> None:
        """Set the min and max clipping values as tuple
        helper function to set the min and max values for clipping of the layer's weights.
        structure of an accepted tuple: (min, max)
        if tuple does not fit above structure, clipping will not be used.
        Also values in tuple must be numerics
        """
        if clipping is None:
            return
        if len(clipping) == 2 and clipping[0] < clipping[1]:
            try:
                float(clipping[0])
                float(clipping[1])
                self._clipping = clipping
            except ValueError:
                raise ValueError("Values must be numerics")
        else:
            raise ValueError("Clipping must be tuple of structure: (min, max)")

    @property
    def probability(self) -> float:
        return self._errorProbability

    @probability.setter
    def probability(self, probability: float) -> None:
        if probability < .0:
            raise ValueError("Probablility cannot be negative")
        self._errorProbability = probability

    @property
    def clipping(self) -> tuple | None:
        return self._clipping

    @clipping.setter
    def clipping(self, clipping: tuple) -> None:
        self._parseClipping(clipping)



from setuptools import setup, find_packages

setup(
    name="NEBULANN",
    version="0.2",
    description="Error injection and quantized training lib for NNs",
    author="Alexander Tepe",
    author_email="alexander.tepe@hotmail.de",
    packages=find_packages(),
    license="MIT",
    install_requires=[
        "setuptools",
        "tensorflow",
        "numpy",
        "python-dotenv"
    ],
)



import unittest
import tensorflow as tf
import numpy as np

from tensorflow._api.v2.errors import InvalidArgumentError

from NEBULA.utils.commons import flipAdjacentBits, flipTensorBits
from NEBULA.utils.helper import binary


class CommonsTest(unittest.TestCase):

    def test_flipTensorBits(self):
        with self.assertRaises(InvalidArgumentError):
            tensor = tf.constant([1.0], dtype=tf.float32)
            tensorFlipped = flipTensorBits(tensor, probability=1.0, dtype=np.float32)
            tf.debugging.assert_equal(tensor, tensorFlipped)

    def test_burstErrorsAreActuallyAdjacent(self):
        number = 12.0
        binString = binary(number)
        numFlipped = flipAdjacentBits(number, 3, 1.0)
        binStringFlipped = binary(numFlipped)
        self.assertEqual(len(binString), len(binStringFlipped))
        diff = list()
        for i in range(len(binString)):
            if binString[i] != binStringFlipped[i]:
                diff.append(i)
        self.assertEqual(len(diff), 3)
        self.assertEqual(diff[-1] - diff[0], 2)



import unittest

from NEBULA.core.trainingInjector import TrainingInjector
from NEBULA.core.trainingInjector import _buildFunctionalModel, _buildSequentialModel
from NEBULA.core.noiseLayer import NoiseLayer

from utils.ModelUtils import ModelUtils


class TrainingInjectorTest(unittest.TestCase):

    def test_buildFuncModelWorks(self):
        model = ModelUtils.getBasicModel()
        nl = NoiseLayer()
        model = _buildFunctionalModel(model, nl, 2)
        self.assertEqual(len(model.layers), 4)
        self.assertTrue("noise_layer" in model.get_layer(index=2).name)

    def test_buildSeqModelWorks(self):
        model = ModelUtils.getSequentialModel()
        nl = NoiseLayer()
        model = _buildSequentialModel(model, nl, 2)
        self.assertEqual(len(model.layers), 4)
        self.assertTrue("noise_layer" in model.get_layer(index=2).name)

    def test_attachFailsWithNegativeIndex(self):
        with self.assertRaises(ValueError):
            ti = TrainingInjector()
            model = ModelUtils.getBasicModel()
            ti.attach(model, -1)



#!/usr/bin/env python3

"""
legacyInjector.py:
    access to the WSA example functions using the injector-wrapper implementation
"""

__author__      = "Alexander Tepe"
__email__       = "alexander.tepe@hotmail.de"
__copyright__   = "Copyright 2024, Planet Earth"

from keras import Model, Layer

from NEBULA.core.baseInjector import BaseInjector
from NEBULA.core.legacy import flip_random_bits_in_model_weights
from NEBULA.utils.logging import getLogger


class LegacyInjector(BaseInjector):
    """Easy access to an error injector using the legacy implementation
    """
    _logger = None
    _check = -1

    def __init__(self, layers: list[Layer], probability=0.01, check=-1) -> None:
        super().__init__(layers, probability)
        self._logger = getLogger(__name__)
        self._check = check

    def injectError(self, model: Model) -> None:
        """calls the og implementation
        This method edits the model inplace.
        """
        self._logger.debug(f"Injecting error with probability of {self._probability}")
        # edit model in place
        flip_random_bits_in_model_weights(model, self._probability, self._check)
